import csv
import functools as ft
from pathlib import Path
from typing import Any


def initialise_csv(fn):

    init = False
    def _init_csv(csv_path, csv_dict):

        with open(csv_path, 'w+', newline='') as f:
            writer = csv.writer(f, delimiter=',')
            writer.writerow(list(csv_dict))

    @ft.wraps(fn)
    def _fn(csv_path, csv_dict):

        nonlocal init

        if not init:
            init = True
            _init_csv(csv_path, csv_dict)

        return fn(csv_path, csv_dict)

    return _fn


@initialise_csv
def write_dict_to_csv(csv_path: Path, csv_dict: dict[str, Any]) -> None:

    with open(csv_path, 'a', newline='') as f:
        writer = csv.writer(f, delimiter=',')
        writer.writerow(list(csv_dict.values()))



from __future__ import annotations

import timeit
from typing import Any, Callable

import jax.numpy as jnp

from gravlax.types import PyTree

from .tree_utils import reduce_pytree, tree_stack


def prepend_to_key(d, prefix):
    return {f'{prefix}{k}': v for k, v in d.items()}


class BatchManager:

    def __init__(self, name: str, n_batches=10):

        self.name = name
        self.n_batches = n_batches

        self.start_time: float
        self._loss_list = []

    @property
    def time_elapsed(self):
        return timeit.default_timer() - self.start_time

    def register_loss(self, loss_dict: dict[str, PyTree]):
        self._loss_list.append(loss_dict)

    def reduce(self, op: Callable[..., PyTree] = jnp.mean, **op_kwargs: dict[str, Any]):
        return prepend_to_key(reduce_pytree(tree_stack(self._loss_list), op, **op_kwargs), prefix=f'{self.name}_')

    def __enter__(self) -> BatchManager:
        self.start_time = timeit.default_timer()
        return self

    def __exit__(self, *args):

        self.time = self.time_elapsed

        if not (n := len(self._loss_list)) == self.n_batches:
            raise ValueError(f'Only recorded {n}/{self.n_batches} batches.')



from typing import Any, Callable, Sequence

import jax
import jax.numpy as jnp
import jax.tree_util as jtu

from .types import PyTree


def tree_stack(trees: Sequence[PyTree]) -> PyTree:
    return jtu.tree_map(lambda *v: jnp.stack(v), *trees)


def tree_unstack(tree: PyTree) -> list[PyTree]:
    leaves, treedef = jtu.tree_flatten(tree)
    return [treedef.unflatten(leaf) for leaf in zip(*leaves, strict=True)]


def reduce_pytree(pytree: PyTree, op: Callable[..., jax.Array] = jnp.mean, op_kwargs: dict[str, Any] | None = None) -> PyTree:

    if not op_kwargs:
        op_kwargs = {}

    return jtu.tree_map(lambda x: op(x, **op_kwargs), pytree)
